/*
 * $Id: auto.c.in,v 1.13 1999/08/18 15:16:20 marcus Exp $
 *
 */

#include "global.h"

#include "config.h"

RCSID("$Id: auto.c.in,v 1.13 1999/08/18 15:16:20 marcus Exp $");
#include "stralloc.h"
#include "pike_macros.h"
#include "object.h"
#include "program.h"
#include "interpret.h"
#include "builtin_functions.h"
#include "error.h"
#include "module_support.h"
#include "operators.h"
#include "mapping.h"

#ifdef HAVE_WINDEF_H
#include <windef.h>
#endif /* HAVE_WINDEF_H */
#ifdef HAVE_WINGDI_H
#include <wingdi.h>
#endif /* HAVE_WINGDI_H */
#ifdef HAVE_GL_GL_H
#include <GL/gl.h>
#endif /* HAVE_GL_GL_H */

#define ZT_INT    1
#define ZT_FLOAT  2
#define ZT_DOUBLE 4
#define ZT_ARRAY  8

struct zvalue4 {
  union {
    GLint i[4];
    GLfloat f[4];
    GLdouble d[4];
  } v;
  int ty;
};

union zvalue16 {
  GLint i[16];
  GLfloat f[16];
  GLdouble d[16];
};

struct zimage {
  GLsizei width, height;
  GLenum format, type;
  GLvoid *pixels;
  int alloc;
};

static int check_234_args(char *func, INT32 args, int mn, int mx, int ty,
			  int rty, struct zvalue4 *ret)
{
  struct svalue *s;
  int i, tt, tr=0;

  if(args==1 && sp[-1].type==T_ARRAY) {
    struct array *a = sp[-1].u.array;
    args = a->size;
    if(args<mn || args>mx)
      error("%s: expected array of size %d-%d.\n", func, mn, mx);
    s = ITEM(a);
  } else {
    if(args<mn || args>mx)
      error("%s: too %s arguments.\n", func, (args<mn? "few":"many"));
    s = sp-args;
  }
  for(i=0; i<args; i++)
    if(!((tt=1<<s[i].type)&ty)) {
      char buf[32];
      int n=0;
      buf[0]='\0';
      for(tt=0; ty; tt++)
	if(ty&(1<<tt)) {
	  ty&=~(1<<tt);
	  if(n++)
	    strcat(buf, " or ");
	  strcat(buf, get_name_of_type(tt));
	}
      error("%s: got %s, expected %s.\n", func,
	    get_name_of_type(s[i].type), buf);
    } else
      tr|=tt;
  if(tr==(BIT_INT|BIT_FLOAT))
    tt=((rty&ZT_DOUBLE)? ZT_DOUBLE:ZT_FLOAT);
  else if(tr==BIT_FLOAT)
    tt=((rty&ZT_FLOAT)? ZT_FLOAT:ZT_DOUBLE);
  else if(tr!=BIT_INT)
    error("Internal error in check_234_args for %s!\n", func);
  else if(rty&ZT_INT)
    tt=ZT_INT;
  else 
    tt=((rty&ZT_DOUBLE)? ZT_DOUBLE:ZT_FLOAT);
  ret->ty=tt;
  for(i=0; i<args; i++) 
    switch(tt)
    {
    case ZT_INT:
      ret->v.i[i]=s[i].u.integer;
      break;
    case ZT_FLOAT:
      ret->v.f[i]=(s[i].type==T_INT?(float)s[i].u.integer:s[i].u.float_number);
      break;      
    case ZT_DOUBLE:
      ret->v.d[i]=(s[i].type==T_INT?(double)s[i].u.integer:
		   s[i].u.float_number);
      break;      
    }
  return args;
}

static int img_inited=0;

static void img_init()
{
  glPixelStorei(GL_PACK_SWAP_BYTES, 0);
  glPixelStorei(GL_PACK_LSB_FIRST, 0);
  glPixelStorei(GL_PACK_ROW_LENGTH, 0);
  glPixelStorei(GL_PACK_SKIP_ROWS, 0);
  glPixelStorei(GL_PACK_SKIP_PIXELS, 0);
  glPixelStorei(GL_PACK_ALIGNMENT, 1);
  glPixelStorei(GL_UNPACK_SWAP_BYTES, 0);
  glPixelStorei(GL_UNPACK_LSB_FIRST, 0);
  glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
  glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
  glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
  img_inited=1;
}

static void check_img_arg(struct svalue *a, struct zimage *img,
			  int arg, const char *func)
{
  struct {
    void *img;
    INT_TYPE w, h;
  } *rgbstor = NULL, *lumstor = NULL, *astor = NULL;
  unsigned char *pp;
  struct { unsigned char r, g, b; } *p1, *p2;
  INT32 cnt;

  static struct program *image_program=NULL;

  if(image_program == NULL) {
    push_string(make_shared_string("Image")); push_int(0);
    SAFE_APPLY_MASTER("resolv",2);
    if (sp[-1].type==T_OBJECT) 
    {
      push_string(make_shared_string("image"));
      f_index(2);
      image_program=program_from_svalue(sp-1);
      pop_stack();
    }
    if(image_program == NULL)
      error("%s: Can't resolve Image.image!\n", func);
  }

  img->alloc = 0;
  if(a->type == T_MAPPING) {
    struct svalue *v;
    if((v=simple_mapping_string_lookup(a->u.mapping, "rgb")))
      if(v->type == T_OBJECT &&
	 (rgbstor = (void *)get_storage(v->u.object, image_program)) != NULL &&
	 rgbstor->img != NULL)
	;
      else
	error("Bad argument %d to %s.\n", arg, func);
    if((v=simple_mapping_string_lookup(a->u.mapping, "luminance")))
      if(v->type == T_OBJECT &&
	 (lumstor = (void *)get_storage(v->u.object, image_program)) != NULL &&
	 lumstor->img != NULL)
	;
      else
	error("Bad argument %d to %s.\n", arg, func);
    if((v=simple_mapping_string_lookup(a->u.mapping, "alpha")))
      if(v->type == T_OBJECT &&
	 (astor = (void *)get_storage(v->u.object, image_program)) != NULL &&
	 astor->img != NULL)
	;
      else
	error("Bad argument %d to %s.\n", arg, func);    
  } else if(a->type == T_OBJECT &&
	    (rgbstor = (void *)get_storage(a->u.object, image_program)) !=
	    NULL && rgbstor->img != NULL)
    ;
  else
    error("Bad argument %d to %s.\n", arg, func);

  if(rgbstor==NULL && lumstor==NULL && astor==NULL)
    error("Too few images specified.\n");

  if(rgbstor!=NULL && lumstor!=NULL)
    error("Can't have both rgb and luminance data.\n");

  if((rgbstor != NULL && astor != NULL &&
      (rgbstor->w != astor->w || rgbstor->h != astor->h)) ||
     (lumstor != NULL && astor != NULL &&
      (lumstor->w != astor->w || lumstor->h != astor->h)))
    error("Alpha channel must have same size as image!\n");

  if(lumstor==NULL && astor==NULL) {
    /* RGB */
    img->width = rgbstor->w;
    img->height = rgbstor->h;
    img->pixels = rgbstor->img;
    img->format = GL_RGB;
    img->type = GL_UNSIGNED_BYTE;
  } else if(rgbstor != NULL) {
    /* RGBA */
    img->width = rgbstor->w;
    img->height = rgbstor->h;
    img->pixels = pp = xalloc(img->width*img->height*4);
    img->format = GL_RGBA;
    img->type = GL_UNSIGNED_BYTE;
    cnt = img->width * img->height;
    p1 = rgbstor->img;
    p2 = astor->img;
    while(cnt--) {
      *pp++ = p1->r;
      *pp++ = p1->g;
      *pp++ = p1++->b;
      *pp++ = p2++->r;
    }
  } else if(astor == NULL) {
    /* Y */
    img->width = lumstor->w;
    img->height = lumstor->h;
    img->pixels = pp = xalloc(img->width*img->height);
    img->format = GL_LUMINANCE;
    img->type = GL_UNSIGNED_BYTE;
    cnt = img->width * img->height;
    p1 = lumstor->img;
    while(cnt--)
      *pp++ = p1++->r;
  } else if(lumstor == NULL) {
    /* A */
    img->width = astor->w;
    img->height = astor->h;
    img->pixels = pp = xalloc(img->width*img->height);
    img->format = GL_ALPHA;
    img->type = GL_UNSIGNED_BYTE;
    cnt = img->width * img->height;
    p1 = astor->img;
    while(cnt--)
      *pp++ = p1++->r;
  } else {
    /* YA */
    img->width = lumstor->w;
    img->height = lumstor->h;
    img->pixels = pp = xalloc(img->width*img->height*2);
    img->format = GL_LUMINANCE_ALPHA;
    img->type = GL_UNSIGNED_BYTE;
    cnt = img->width * img->height;
    p1 = lumstor->img;
    p2 = astor->img;
    while(cnt--) {
      *pp++ = p1++->r;
      *pp++ = p2++->r;
    }
  }
  if(!img_inited)
    img_init();
}

static void release_img(struct zimage *img)
{
  if(img->alloc)
    free(img->pixels);
}

static int check_1n_args(char *func, INT32 args, int ty, int rty,
			 union zvalue16 *ret)
{
  int i, tt, tr=0, arr=0;
  struct svalue *s;

  if(args!=1)
    error("%s: too %s arguments.\n", func, (args<1? "few":"many"));
  
  if(sp[-1].type==T_ARRAY) {
    struct array *a = sp[-1].u.array;
    args = a->size;
    if(args>16)
      error("%s: array too large.\n", func);
    arr++;
    s = ITEM(a);
  } else
    s = sp-args;

  for(i=0; i<args; i++)
    if(!((tt=1<<s[i].type)&ty)) {
      char buf[32];
      int n=0;
      buf[0]='\0';
      for(tt=0; ty; tt++)
	if(ty&(1<<tt)) {
	  ty&=~(1<<tt);
	  if(n++)
	    strcat(buf, " or ");
	  strcat(buf, get_name_of_type(tt));
	}
      error("%s: got %s, expected %s.\n", func,
	    get_name_of_type(s[i].type), buf);
    } else
      tr|=tt;
  if(tr==(BIT_INT|BIT_FLOAT))
    tt=((rty&ZT_DOUBLE)? ZT_DOUBLE:ZT_FLOAT);
  else if(tr==BIT_FLOAT)
    tt=((rty&ZT_FLOAT)? ZT_FLOAT:ZT_DOUBLE);
  else if(tr!=BIT_INT)
    error("Internal error in check_1n_args for %s!\n", func);
  else if(rty&ZT_INT)
    tt=ZT_INT;
  else 
    tt=((rty&ZT_DOUBLE)? ZT_DOUBLE:ZT_FLOAT);

  for(i=0; i<args; i++) 
    switch(tt)
    {
    case ZT_INT:
      ret->i[i]=s[i].u.integer;
      break;
    case ZT_FLOAT:
      ret->f[i]=(s[i].type==T_INT?(float)s[i].u.integer:s[i].u.float_number);
      break;      
    case ZT_DOUBLE:
      ret->d[i]=(s[i].type==T_INT?(double)s[i].u.integer:s[i].u.float_number);
      break;      
    }  

  return tt|(arr?ZT_ARRAY:0);
}

#define glClearColor3v(f) glClearColor((f)[0], (f)[1], (f)[2], 0.0)
#define glClearColor4v(f) glClearColor((f)[0], (f)[1], (f)[2], (f)[3])
#define glClearAccum3v(f) glClearAccum((f)[0], (f)[1], (f)[2], 0.0)
#define glClearAccum4v(f) glClearAccum((f)[0], (f)[1], (f)[2], (f)[3])
#define glEvalPoint1v(i) glEvalPoint1((i)[0]);
#define glEvalPoint2v(i) glEvalPoint2((i)[0], (i)[1]);

static void pre_init()
{
}

static void post_init()
{
}

static void f_glAccum(INT32 args)
{
  INT32 arg1;
  float arg2;

  check_all_args("glAccum", args, BIT_INT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.float_number;

  glAccum(arg1,arg2);
  pop_n_elems(args);
}

static void f_glAlphaFunc(INT32 args)
{
  INT32 arg1;
  float arg2;

  check_all_args("glAlphaFunc", args, BIT_INT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.float_number;

  glAlphaFunc(arg1,arg2);
  pop_n_elems(args);
}

static void f_glArrayElement(INT32 args)
{
  INT32 arg1;

  check_all_args("glArrayElement", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glArrayElement(arg1);
  pop_n_elems(args);
}

static void f_glBegin(INT32 args)
{
  INT32 arg1;

  check_all_args("glBegin", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glBegin(arg1);
  pop_n_elems(args);
}

static void f_glBindTexture(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glBindTexture", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glBindTexture(arg1,arg2);
  pop_n_elems(args);
}

static void f_glBlendFunc(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glBlendFunc", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glBlendFunc(arg1,arg2);
  pop_n_elems(args);
}

static void f_glCallList(INT32 args)
{
  INT32 arg1;

  check_all_args("glCallList", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glCallList(arg1);
  pop_n_elems(args);
}

static void f_glClear(INT32 args)
{
  INT32 arg1;

  check_all_args("glClear", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glClear(arg1);
  pop_n_elems(args);
}

static void f_glClearAccum(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glClearAccum", args-0, 3, 4, BIT_FLOAT, ZT_FLOAT, &zv4);

  check_all_args("glClearAccum", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(r234) {
    case 3: glClearAccum3v(zv4.v.f); break;
    case 4: glClearAccum4v(zv4.v.f); break;
  }
  pop_n_elems(args);
}

static void f_glClearColor(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glClearColor", args-0, 3, 4, BIT_FLOAT, ZT_FLOAT, &zv4);

  check_all_args("glClearColor", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(r234) {
    case 3: glClearColor3v(zv4.v.f); break;
    case 4: glClearColor4v(zv4.v.f); break;
  }
  pop_n_elems(args);
}

static void f_glClearDepth(INT32 args)
{
  double arg1;

  check_all_args("glClearDepth", args, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;

  glClearDepth(arg1);
  pop_n_elems(args);
}

static void f_glClearIndex(INT32 args)
{
  float arg1;

  check_all_args("glClearIndex", args, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;

  glClearIndex(arg1);
  pop_n_elems(args);
}

static void f_glClearStencil(INT32 args)
{
  INT32 arg1;

  check_all_args("glClearStencil", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glClearStencil(arg1);
  pop_n_elems(args);
}

static void f_glClipPlane(INT32 args)
{
  INT32 arg1;
  struct zvalue4 zv4;

  check_234_args("glClipPlane", args-1, 4, 4, BIT_FLOAT, ZT_DOUBLE, &zv4);

  check_all_args("glClipPlane", args, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;

  glClipPlane(arg1,zv4.v.d);
  pop_n_elems(args);
}

static void f_glColor(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glColor", args-0, 3, 4, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glColor", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_INT: switch(r234) {
      case 3: glColor3iv(zv4.v.i); break;
      case 4: glColor4iv(zv4.v.i); break;
    }
    break;
    case ZT_FLOAT: switch(r234) {
      case 3: glColor3fv(zv4.v.f); break;
      case 4: glColor4fv(zv4.v.f); break;
    }
    break;
    case ZT_DOUBLE: switch(r234) {
      case 3: glColor3dv(zv4.v.d); break;
      case 4: glColor4dv(zv4.v.d); break;
    }
    break;
  }
  pop_n_elems(args);
}

static void f_glColorMask(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;

  check_all_args("glColorMask", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;

  glColorMask(arg1,arg2,arg3,arg4);
  pop_n_elems(args);
}

static void f_glColorMaterial(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glColorMaterial", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glColorMaterial(arg1,arg2);
  pop_n_elems(args);
}

static void f_glCopyPixels(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  INT32 arg5;

  check_all_args("glCopyPixels", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  arg5=sp[4-args].u.integer;

  glCopyPixels(arg1,arg2,arg3,arg4,arg5);
  pop_n_elems(args);
}

static void f_glCopyTexImage1D(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  INT32 arg5;
  INT32 arg6;
  INT32 arg7;

  check_all_args("glCopyTexImage1D", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  arg5=sp[4-args].u.integer;
  arg6=sp[5-args].u.integer;
  arg7=sp[6-args].u.integer;

  glCopyTexImage1D(arg1,arg2,arg3,arg4,arg5,arg6,arg7);
  pop_n_elems(args);
}

static void f_glCopyTexImage2D(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  INT32 arg5;
  INT32 arg6;
  INT32 arg7;
  INT32 arg8;

  check_all_args("glCopyTexImage2D", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  arg5=sp[4-args].u.integer;
  arg6=sp[5-args].u.integer;
  arg7=sp[6-args].u.integer;
  arg8=sp[7-args].u.integer;

  glCopyTexImage2D(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  pop_n_elems(args);
}

static void f_glCopyTexSubImage1D(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  INT32 arg5;
  INT32 arg6;

  check_all_args("glCopyTexSubImage1D", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  arg5=sp[4-args].u.integer;
  arg6=sp[5-args].u.integer;

  glCopyTexSubImage1D(arg1,arg2,arg3,arg4,arg5,arg6);
  pop_n_elems(args);
}

static void f_glCopyTexSubImage2D(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  INT32 arg5;
  INT32 arg6;
  INT32 arg7;
  INT32 arg8;

  check_all_args("glCopyTexSubImage2D", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  arg5=sp[4-args].u.integer;
  arg6=sp[5-args].u.integer;
  arg7=sp[6-args].u.integer;
  arg8=sp[7-args].u.integer;

  glCopyTexSubImage2D(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8);
  pop_n_elems(args);
}

static void f_glCullFace(INT32 args)
{
  INT32 arg1;

  check_all_args("glCullFace", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glCullFace(arg1);
  pop_n_elems(args);
}

static void f_glDeleteLists(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glDeleteLists", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glDeleteLists(arg1,arg2);
  pop_n_elems(args);
}

static void f_glDepthFunc(INT32 args)
{
  INT32 arg1;

  check_all_args("glDepthFunc", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glDepthFunc(arg1);
  pop_n_elems(args);
}

static void f_glDepthMask(INT32 args)
{
  INT32 arg1;

  check_all_args("glDepthMask", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glDepthMask(arg1);
  pop_n_elems(args);
}

static void f_glDepthRange(INT32 args)
{
  double arg1;
  double arg2;

  check_all_args("glDepthRange", args, BIT_FLOAT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;
  arg2=sp[1-args].u.float_number;

  glDepthRange(arg1,arg2);
  pop_n_elems(args);
}

static void f_glDisable(INT32 args)
{
  INT32 arg1;

  check_all_args("glDisable", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glDisable(arg1);
  pop_n_elems(args);
}

static void f_glDisableClientState(INT32 args)
{
  INT32 arg1;

  check_all_args("glDisableClientState", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glDisableClientState(arg1);
  pop_n_elems(args);
}

static void f_glDrawArrays(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;

  check_all_args("glDrawArrays", args, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;

  glDrawArrays(arg1,arg2,arg3);
  pop_n_elems(args);
}

static void f_glDrawBuffer(INT32 args)
{
  INT32 arg1;

  check_all_args("glDrawBuffer", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glDrawBuffer(arg1);
  pop_n_elems(args);
}

static void f_glDrawPixels(INT32 args)
{
  struct zimage img;

  check_all_args("glDrawPixels", args, BIT_OBJECT|BIT_MAPPING, 0);

  check_img_arg(&sp[0-args], &img, 1, "glDrawPixels");

  glDrawPixels(img.width,img.height,img.format,img.type,img.pixels);
  release_img(&img);
  pop_n_elems(args);
}

static void f_glEdgeFlag(INT32 args)
{
  INT32 arg1;

  check_all_args("glEdgeFlag", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glEdgeFlag(arg1);
  pop_n_elems(args);
}

static void f_glEnable(INT32 args)
{
  INT32 arg1;

  check_all_args("glEnable", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glEnable(arg1);
  pop_n_elems(args);
}

static void f_glEnableClientState(INT32 args)
{
  INT32 arg1;

  check_all_args("glEnableClientState", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glEnableClientState(arg1);
  pop_n_elems(args);
}

static void f_glEnd(INT32 args)
{
  glEnd();
  pop_n_elems(args);
}

static void f_glEndList(INT32 args)
{
  glEndList();
  pop_n_elems(args);
}

static void f_glEvalCoord(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glEvalCoord", args-0, 1, 2, BIT_FLOAT|BIT_INT, ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glEvalCoord", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_FLOAT: switch(r234) {
      case 1: glEvalCoord1fv(zv4.v.f); break;
      case 2: glEvalCoord2fv(zv4.v.f); break;
    }
    break;
    case ZT_DOUBLE: switch(r234) {
      case 1: glEvalCoord1dv(zv4.v.d); break;
      case 2: glEvalCoord2dv(zv4.v.d); break;
    }
    break;
  }
  pop_n_elems(args);
}

static void f_glEvalPoint(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glEvalPoint", args-0, 1, 2, BIT_INT, ZT_INT, &zv4);

  check_all_args("glEvalPoint", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(r234) {
    case 1: glEvalPoint1v(zv4.v.i); break;
    case 2: glEvalPoint2v(zv4.v.i); break;
  }
  pop_n_elems(args);
}

static void f_glFinish(INT32 args)
{
  glFinish();
  pop_n_elems(args);
}

static void f_glFlush(INT32 args)
{
  glFlush();
  pop_n_elems(args);
}

static void f_glFog(INT32 args)
{
  INT32 arg1;
  union zvalue16 zv16;

  int r1n=check_1n_args("glFog", args-1, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT, &zv16);

  check_all_args("glFog", args, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;

  switch(r1n) {
    case ZT_INT: glFogi(arg1,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glFogiv(arg1,zv16.i); break;
    case ZT_FLOAT: glFogf(arg1,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glFogfv(arg1,zv16.f); break;
  }
  pop_n_elems(args);
}

static void f_glFrontFace(INT32 args)
{
  INT32 arg1;

  check_all_args("glFrontFace", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glFrontFace(arg1);
  pop_n_elems(args);
}

static void f_glFrustum(INT32 args)
{
  double arg1;
  double arg2;
  double arg3;
  double arg4;
  double arg5;
  double arg6;

  check_all_args("glFrustum", args, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;
  arg2=sp[1-args].u.float_number;
  arg3=sp[2-args].u.float_number;
  arg4=sp[3-args].u.float_number;
  arg5=sp[4-args].u.float_number;
  arg6=sp[5-args].u.float_number;

  glFrustum(arg1,arg2,arg3,arg4,arg5,arg6);
  pop_n_elems(args);
}

static void f_glGenLists(INT32 args)
{
  INT32 res;
  INT32 arg1;

  check_all_args("glGenLists", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  res=glGenLists(arg1);
  pop_n_elems(args);
  push_int(res);
}

static void f_glGetError(INT32 args)
{
  INT32 res;
  res=glGetError();
  pop_n_elems(args);
  push_int(res);
}

static void f_glGetString(INT32 args)
{
  const GLubyte * res;
  INT32 arg1;

  check_all_args("glGetString", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  res=glGetString(arg1);
  pop_n_elems(args);
  push_text((char *)res);
}

static void f_glHint(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glHint", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glHint(arg1,arg2);
  pop_n_elems(args);
}

static void f_glIndex(INT32 args)
{

  check_all_args("glIndex", args, BIT_FLOAT|BIT_INT, 0);
  switch(sp[0-args].type) {
    case T_FLOAT: glIndexf(sp[0-args].u.float_number); break;
    case T_INT: glIndexi(sp[0-args].u.integer); break;
  }
  pop_n_elems(args);
}

static void f_glIndexMask(INT32 args)
{
  INT32 arg1;

  check_all_args("glIndexMask", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glIndexMask(arg1);
  pop_n_elems(args);
}

static void f_glInitNames(INT32 args)
{
  glInitNames();
  pop_n_elems(args);
}

static void f_glIsEnabled(INT32 args)
{
  INT32 res;
  INT32 arg1;

  check_all_args("glIsEnabled", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  res=glIsEnabled(arg1);
  pop_n_elems(args);
  push_int(res);
}

static void f_glIsList(INT32 args)
{
  INT32 res;
  INT32 arg1;

  check_all_args("glIsList", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  res=glIsList(arg1);
  pop_n_elems(args);
  push_int(res);
}

static void f_glIsTexture(INT32 args)
{
  INT32 res;
  INT32 arg1;

  check_all_args("glIsTexture", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  res=glIsTexture(arg1);
  pop_n_elems(args);
  push_int(res);
}

static void f_glLight(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  union zvalue16 zv16;

  int r1n=check_1n_args("glLight", args-2, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT, &zv16);

  check_all_args("glLight", args, BIT_INT, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  switch(r1n) {
    case ZT_INT: glLighti(arg1,arg2,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glLightiv(arg1,arg2,zv16.i); break;
    case ZT_FLOAT: glLightf(arg1,arg2,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glLightfv(arg1,arg2,zv16.f); break;
  }
  pop_n_elems(args);
}

static void f_glLightModel(INT32 args)
{
  INT32 arg1;
  union zvalue16 zv16;

  int r1n=check_1n_args("glLightModel", args-1, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT, &zv16);

  check_all_args("glLightModel", args, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;

  switch(r1n) {
    case ZT_INT: glLightModeli(arg1,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glLightModeliv(arg1,zv16.i); break;
    case ZT_FLOAT: glLightModelf(arg1,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glLightModelfv(arg1,zv16.f); break;
  }
  pop_n_elems(args);
}

static void f_glLineStipple(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glLineStipple", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glLineStipple(arg1,arg2);
  pop_n_elems(args);
}

static void f_glLineWidth(INT32 args)
{
  float arg1;

  check_all_args("glLineWidth", args, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;

  glLineWidth(arg1);
  pop_n_elems(args);
}

static void f_glListBase(INT32 args)
{
  INT32 arg1;

  check_all_args("glListBase", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glListBase(arg1);
  pop_n_elems(args);
}

static void f_glLoadIdentity(INT32 args)
{
  glLoadIdentity();
  pop_n_elems(args);
}

static void f_glLoadMatrix(INT32 args)
{
  union zvalue16 zv16;
  int r1n;

  check_all_args("glLoadMatrix", args, BIT_ARRAY, 0);

  if(sp[0-args].u.array->size != 16)
    error("glLoadMatrix: Array length is wrong (is %d, should be 16)\n", sp[0-args].u.array->size);

  r1n=check_1n_args("glLoadMatrix", args-0, BIT_FLOAT|BIT_INT, ZT_FLOAT|ZT_DOUBLE, &zv16);

  switch(r1n) {
    case ZT_ARRAY|ZT_FLOAT: glLoadMatrixf(zv16.f); break;
    case ZT_ARRAY|ZT_DOUBLE: glLoadMatrixd(zv16.d); break;
  }
  pop_n_elems(args);
}

static void f_glLoadName(INT32 args)
{
  INT32 arg1;

  check_all_args("glLoadName", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glLoadName(arg1);
  pop_n_elems(args);
}

static void f_glLogicOp(INT32 args)
{
  INT32 arg1;

  check_all_args("glLogicOp", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glLogicOp(arg1);
  pop_n_elems(args);
}

static void f_glMaterial(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  union zvalue16 zv16;

  int r1n=check_1n_args("glMaterial", args-2, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT, &zv16);

  check_all_args("glMaterial", args, BIT_INT, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  switch(r1n) {
    case ZT_INT: glMateriali(arg1,arg2,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glMaterialiv(arg1,arg2,zv16.i); break;
    case ZT_FLOAT: glMaterialf(arg1,arg2,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glMaterialfv(arg1,arg2,zv16.f); break;
  }
  pop_n_elems(args);
}

static void f_glMatrixMode(INT32 args)
{
  INT32 arg1;

  check_all_args("glMatrixMode", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glMatrixMode(arg1);
  pop_n_elems(args);
}

static void f_glMultMatrix(INT32 args)
{
  union zvalue16 zv16;
  int r1n;

  check_all_args("glMultMatrix", args, BIT_ARRAY, 0);

  if(sp[0-args].u.array->size != 16)
    error("glMultMatrix: Array length is wrong (is %d, should be 16)\n", sp[0-args].u.array->size);

  r1n=check_1n_args("glMultMatrix", args-0, BIT_FLOAT|BIT_INT, ZT_FLOAT|ZT_DOUBLE, &zv16);

  switch(r1n) {
    case ZT_ARRAY|ZT_FLOAT: glMultMatrixf(zv16.f); break;
    case ZT_ARRAY|ZT_DOUBLE: glMultMatrixd(zv16.d); break;
  }
  pop_n_elems(args);
}

static void f_glNewList(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glNewList", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glNewList(arg1,arg2);
  pop_n_elems(args);
}

static void f_glNormal(INT32 args)
{
  struct zvalue4 zv4;

  check_234_args("glNormal", args-0, 3, 3, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glNormal", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_INT: glNormal3iv(zv4.v.i); break;
    case ZT_FLOAT: glNormal3fv(zv4.v.f); break;
    case ZT_DOUBLE: glNormal3dv(zv4.v.d); break;
  }
  pop_n_elems(args);
}

static void f_glOrtho(INT32 args)
{
  double arg1;
  double arg2;
  double arg3;
  double arg4;
  double arg5;
  double arg6;

  check_all_args("glOrtho", args, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;
  arg2=sp[1-args].u.float_number;
  arg3=sp[2-args].u.float_number;
  arg4=sp[3-args].u.float_number;
  arg5=sp[4-args].u.float_number;
  arg6=sp[5-args].u.float_number;

  glOrtho(arg1,arg2,arg3,arg4,arg5,arg6);
  pop_n_elems(args);
}

static void f_glPassThrough(INT32 args)
{
  float arg1;

  check_all_args("glPassThrough", args, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;

  glPassThrough(arg1);
  pop_n_elems(args);
}

static void f_glPixelZoom(INT32 args)
{
  float arg1;
  float arg2;

  check_all_args("glPixelZoom", args, BIT_FLOAT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;
  arg2=sp[1-args].u.float_number;

  glPixelZoom(arg1,arg2);
  pop_n_elems(args);
}

static void f_glPointSize(INT32 args)
{
  float arg1;

  check_all_args("glPointSize", args, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;

  glPointSize(arg1);
  pop_n_elems(args);
}

static void f_glPolygonMode(INT32 args)
{
  INT32 arg1;
  INT32 arg2;

  check_all_args("glPolygonMode", args, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  glPolygonMode(arg1,arg2);
  pop_n_elems(args);
}

static void f_glPolygonOffset(INT32 args)
{
  float arg1;
  float arg2;

  check_all_args("glPolygonOffset", args, BIT_FLOAT, BIT_FLOAT, 0);

  arg1=sp[0-args].u.float_number;
  arg2=sp[1-args].u.float_number;

  glPolygonOffset(arg1,arg2);
  pop_n_elems(args);
}

static void f_glPopAttrib(INT32 args)
{
  glPopAttrib();
  pop_n_elems(args);
}

static void f_glPopClientAttrib(INT32 args)
{
  glPopClientAttrib();
  pop_n_elems(args);
}

static void f_glPopMatrix(INT32 args)
{
  glPopMatrix();
  pop_n_elems(args);
}

static void f_glPopName(INT32 args)
{
  glPopName();
  pop_n_elems(args);
}

static void f_glPushAttrib(INT32 args)
{
  INT32 arg1;

  check_all_args("glPushAttrib", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glPushAttrib(arg1);
  pop_n_elems(args);
}

static void f_glPushClientAttrib(INT32 args)
{
  INT32 arg1;

  check_all_args("glPushClientAttrib", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glPushClientAttrib(arg1);
  pop_n_elems(args);
}

static void f_glPushMatrix(INT32 args)
{
  glPushMatrix();
  pop_n_elems(args);
}

static void f_glPushName(INT32 args)
{
  INT32 arg1;

  check_all_args("glPushName", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glPushName(arg1);
  pop_n_elems(args);
}

static void f_glRasterPos(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glRasterPos", args-0, 2, 4, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glRasterPos", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_INT: switch(r234) {
      case 2: glRasterPos2iv(zv4.v.i); break;
      case 3: glRasterPos3iv(zv4.v.i); break;
      case 4: glRasterPos4iv(zv4.v.i); break;
    }
    break;
    case ZT_FLOAT: switch(r234) {
      case 2: glRasterPos2fv(zv4.v.f); break;
      case 3: glRasterPos3fv(zv4.v.f); break;
      case 4: glRasterPos4fv(zv4.v.f); break;
    }
    break;
    case ZT_DOUBLE: switch(r234) {
      case 2: glRasterPos2dv(zv4.v.d); break;
      case 3: glRasterPos3dv(zv4.v.d); break;
      case 4: glRasterPos4dv(zv4.v.d); break;
    }
    break;
  }
  pop_n_elems(args);
}

static void f_glReadBuffer(INT32 args)
{
  INT32 arg1;

  check_all_args("glReadBuffer", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glReadBuffer(arg1);
  pop_n_elems(args);
}

static void f_glRenderMode(INT32 args)
{
  INT32 arg1;

  check_all_args("glRenderMode", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glRenderMode(arg1);
  pop_n_elems(args);
}

static void f_glRotate(INT32 args)
{
  struct zvalue4 zv4;

  check_234_args("glRotate", args-0, 4, 4, BIT_FLOAT|BIT_INT, ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glRotate", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_FLOAT: glRotatef(zv4.v.f[0],zv4.v.f[1],zv4.v.f[2],zv4.v.f[3]); break;
    case ZT_DOUBLE: glRotated(zv4.v.d[0],zv4.v.d[1],zv4.v.d[2],zv4.v.d[3]); break;
  }
  pop_n_elems(args);
}

static void f_glScale(INT32 args)
{
  struct zvalue4 zv4;

  check_234_args("glScale", args-0, 3, 3, BIT_FLOAT|BIT_INT, ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glScale", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_FLOAT: glScalef(zv4.v.f[0],zv4.v.f[1],zv4.v.f[2]); break;
    case ZT_DOUBLE: glScaled(zv4.v.d[0],zv4.v.d[1],zv4.v.d[2]); break;
  }
  pop_n_elems(args);
}

static void f_glScissor(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;

  check_all_args("glScissor", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;

  glScissor(arg1,arg2,arg3,arg4);
  pop_n_elems(args);
}

static void f_glShadeModel(INT32 args)
{
  INT32 arg1;

  check_all_args("glShadeModel", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glShadeModel(arg1);
  pop_n_elems(args);
}

static void f_glStencilFunc(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;

  check_all_args("glStencilFunc", args, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;

  glStencilFunc(arg1,arg2,arg3);
  pop_n_elems(args);
}

static void f_glStencilMask(INT32 args)
{
  INT32 arg1;

  check_all_args("glStencilMask", args, BIT_INT, 0);

  arg1=sp[0-args].u.integer;

  glStencilMask(arg1);
  pop_n_elems(args);
}

static void f_glStencilOp(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;

  check_all_args("glStencilOp", args, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;

  glStencilOp(arg1,arg2,arg3);
  pop_n_elems(args);
}

static void f_glTexCoord(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glTexCoord", args-0, 1, 4, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glTexCoord", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_INT: switch(r234) {
      case 1: glTexCoord1iv(zv4.v.i); break;
      case 2: glTexCoord2iv(zv4.v.i); break;
      case 3: glTexCoord3iv(zv4.v.i); break;
      case 4: glTexCoord4iv(zv4.v.i); break;
    }
    break;
    case ZT_FLOAT: switch(r234) {
      case 1: glTexCoord1fv(zv4.v.f); break;
      case 2: glTexCoord2fv(zv4.v.f); break;
      case 3: glTexCoord3fv(zv4.v.f); break;
      case 4: glTexCoord4fv(zv4.v.f); break;
    }
    break;
    case ZT_DOUBLE: switch(r234) {
      case 1: glTexCoord1dv(zv4.v.d); break;
      case 2: glTexCoord2dv(zv4.v.d); break;
      case 3: glTexCoord3dv(zv4.v.d); break;
      case 4: glTexCoord4dv(zv4.v.d); break;
    }
    break;
  }
  pop_n_elems(args);
}

static void f_glTexEnv(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  union zvalue16 zv16;

  int r1n=check_1n_args("glTexEnv", args-2, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT, &zv16);

  check_all_args("glTexEnv", args, BIT_INT, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  switch(r1n) {
    case ZT_INT: glTexEnvi(arg1,arg2,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glTexEnviv(arg1,arg2,zv16.i); break;
    case ZT_FLOAT: glTexEnvf(arg1,arg2,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glTexEnvfv(arg1,arg2,zv16.f); break;
  }
  pop_n_elems(args);
}

static void f_glTexGen(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  union zvalue16 zv16;

  int r1n=check_1n_args("glTexGen", args-2, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT|ZT_DOUBLE, &zv16);

  check_all_args("glTexGen", args, BIT_INT, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  switch(r1n) {
    case ZT_INT: glTexGeni(arg1,arg2,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glTexGeniv(arg1,arg2,zv16.i); break;
    case ZT_FLOAT: glTexGenf(arg1,arg2,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glTexGenfv(arg1,arg2,zv16.f); break;
    case ZT_DOUBLE: glTexGend(arg1,arg2,zv16.d[0]); break;
    case ZT_ARRAY|ZT_DOUBLE: glTexGendv(arg1,arg2,zv16.d); break;
  }
  pop_n_elems(args);
}

static void f_glTexImage2D(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  struct zimage img;

  check_all_args("glTexImage2D", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_OBJECT|BIT_MAPPING, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  check_img_arg(&sp[4-args], &img, 5, "glTexImage2D");

  glTexImage2D(arg1,arg2,arg3,img.width,img.height,arg4,img.format,img.type,img.pixels);
  release_img(&img);
  pop_n_elems(args);
}

static void f_glTexParameter(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  union zvalue16 zv16;

  int r1n=check_1n_args("glTexParameter", args-2, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT, &zv16);

  check_all_args("glTexParameter", args, BIT_INT, BIT_INT, BIT_MANY|BIT_MIXED|BIT_VOID, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;

  switch(r1n) {
    case ZT_INT: glTexParameteri(arg1,arg2,zv16.i[0]); break;
    case ZT_ARRAY|ZT_INT: glTexParameteriv(arg1,arg2,zv16.i); break;
    case ZT_FLOAT: glTexParameterf(arg1,arg2,zv16.f[0]); break;
    case ZT_ARRAY|ZT_FLOAT: glTexParameterfv(arg1,arg2,zv16.f); break;
  }
  pop_n_elems(args);
}

static void f_glTexSubImage2D(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;
  struct zimage img;

  check_all_args("glTexSubImage2D", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, BIT_OBJECT|BIT_MAPPING, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;
  check_img_arg(&sp[4-args], &img, 5, "glTexSubImage2D");

  glTexSubImage2D(arg1,arg2,arg3,arg4,img.width,img.height,img.format,img.type,img.pixels);
  release_img(&img);
  pop_n_elems(args);
}

static void f_glTranslate(INT32 args)
{
  struct zvalue4 zv4;

  check_234_args("glTranslate", args-0, 3, 3, BIT_FLOAT|BIT_INT, ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glTranslate", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_FLOAT: glTranslatef(zv4.v.f[0],zv4.v.f[1],zv4.v.f[2]); break;
    case ZT_DOUBLE: glTranslated(zv4.v.d[0],zv4.v.d[1],zv4.v.d[2]); break;
  }
  pop_n_elems(args);
}

static void f_glVertex(INT32 args)
{
  struct zvalue4 zv4;

  int r234=check_234_args("glVertex", args-0, 2, 4, BIT_FLOAT|BIT_INT, ZT_INT|ZT_FLOAT|ZT_DOUBLE, &zv4);

  check_all_args("glVertex", args, BIT_MANY|BIT_MIXED|BIT_VOID, 0);
  switch(zv4.ty) {
    case ZT_INT: switch(r234) {
      case 2: glVertex2iv(zv4.v.i); break;
      case 3: glVertex3iv(zv4.v.i); break;
      case 4: glVertex4iv(zv4.v.i); break;
    }
    break;
    case ZT_FLOAT: switch(r234) {
      case 2: glVertex2fv(zv4.v.f); break;
      case 3: glVertex3fv(zv4.v.f); break;
      case 4: glVertex4fv(zv4.v.f); break;
    }
    break;
    case ZT_DOUBLE: switch(r234) {
      case 2: glVertex2dv(zv4.v.d); break;
      case 3: glVertex3dv(zv4.v.d); break;
      case 4: glVertex4dv(zv4.v.d); break;
    }
    break;
  }
  pop_n_elems(args);
}

static void f_glViewport(INT32 args)
{
  INT32 arg1;
  INT32 arg2;
  INT32 arg3;
  INT32 arg4;

  check_all_args("glViewport", args, BIT_INT, BIT_INT, BIT_INT, BIT_INT, 0);

  arg1=sp[0-args].u.integer;
  arg2=sp[1-args].u.integer;
  arg3=sp[2-args].u.integer;
  arg4=sp[3-args].u.integer;

  glViewport(arg1,arg2,arg3,arg4);
  pop_n_elems(args);
}

void GL_add_auto_funcs()
{
  pre_init();
  add_function_constant("glAccum", f_glAccum,
			"function(int,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glAlphaFunc", f_glAlphaFunc,
			"function(int,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glArrayElement", f_glArrayElement,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glBegin", f_glBegin,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glBindTexture", f_glBindTexture,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glBlendFunc", f_glBlendFunc,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCallList", f_glCallList,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClear", f_glClear,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClearAccum", f_glClearAccum,
			"function(float|array(float),float|void,float|void,float|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClearColor", f_glClearColor,
			"function(float|array(float),float|void,float|void,float|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClearDepth", f_glClearDepth,
			"function(float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClearIndex", f_glClearIndex,
			"function(float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClearStencil", f_glClearStencil,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glClipPlane", f_glClipPlane,
			"function(int,float|array(float),float|void,float|void,float|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glColor", f_glColor,
			"function(float|int|array(float|int),float|int|void,float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glColorMask", f_glColorMask,
			"function(int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glColorMaterial", f_glColorMaterial,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCopyPixels", f_glCopyPixels,
			"function(int,int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCopyTexImage1D", f_glCopyTexImage1D,
			"function(int,int,int,int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCopyTexImage2D", f_glCopyTexImage2D,
			"function(int,int,int,int,int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCopyTexSubImage1D", f_glCopyTexSubImage1D,
			"function(int,int,int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCopyTexSubImage2D", f_glCopyTexSubImage2D,
			"function(int,int,int,int,int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glCullFace", f_glCullFace,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDeleteLists", f_glDeleteLists,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDepthFunc", f_glDepthFunc,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDepthMask", f_glDepthMask,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDepthRange", f_glDepthRange,
			"function(float,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDisable", f_glDisable,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDisableClientState", f_glDisableClientState,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDrawArrays", f_glDrawArrays,
			"function(int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDrawBuffer", f_glDrawBuffer,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glDrawPixels", f_glDrawPixels,
			"function(object|mapping(string:object):void)", OPT_SIDE_EFFECT);
  add_function_constant("glEdgeFlag", f_glEdgeFlag,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glEnable", f_glEnable,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glEnableClientState", f_glEnableClientState,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glEnd", f_glEnd,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glEndList", f_glEndList,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glEvalCoord", f_glEvalCoord,
			"function(float|int|array(float|int),float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glEvalPoint", f_glEvalPoint,
			"function(int|array(int),int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glFinish", f_glFinish,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glFlush", f_glFlush,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glFog", f_glFog,
			"function(int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glFrontFace", f_glFrontFace,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glFrustum", f_glFrustum,
			"function(float,float,float,float,float,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glGenLists", f_glGenLists,
			"function(int:int)", OPT_SIDE_EFFECT);
  add_function_constant("glGetError", f_glGetError,
			"function(:int)", OPT_SIDE_EFFECT);
  add_function_constant("glGetString", f_glGetString,
			"function(int:string)", OPT_SIDE_EFFECT);
  add_function_constant("glHint", f_glHint,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glIndex", f_glIndex,
			"function(float|int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glIndexMask", f_glIndexMask,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glInitNames", f_glInitNames,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glIsEnabled", f_glIsEnabled,
			"function(int:int)", OPT_SIDE_EFFECT);
  add_function_constant("glIsList", f_glIsList,
			"function(int:int)", OPT_SIDE_EFFECT);
  add_function_constant("glIsTexture", f_glIsTexture,
			"function(int:int)", OPT_SIDE_EFFECT);
  add_function_constant("glLight", f_glLight,
			"function(int,int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glLightModel", f_glLightModel,
			"function(int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glLineStipple", f_glLineStipple,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glLineWidth", f_glLineWidth,
			"function(float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glListBase", f_glListBase,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glLoadIdentity", f_glLoadIdentity,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glLoadMatrix", f_glLoadMatrix,
			"function(array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glLoadName", f_glLoadName,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glLogicOp", f_glLogicOp,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glMaterial", f_glMaterial,
			"function(int,int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glMatrixMode", f_glMatrixMode,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glMultMatrix", f_glMultMatrix,
			"function(array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glNewList", f_glNewList,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glNormal", f_glNormal,
			"function(float|int|array(float|int),float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glOrtho", f_glOrtho,
			"function(float,float,float,float,float,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPassThrough", f_glPassThrough,
			"function(float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPixelZoom", f_glPixelZoom,
			"function(float,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPointSize", f_glPointSize,
			"function(float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPolygonMode", f_glPolygonMode,
			"function(int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPolygonOffset", f_glPolygonOffset,
			"function(float,float:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPopAttrib", f_glPopAttrib,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPopClientAttrib", f_glPopClientAttrib,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPopMatrix", f_glPopMatrix,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPopName", f_glPopName,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPushAttrib", f_glPushAttrib,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPushClientAttrib", f_glPushClientAttrib,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPushMatrix", f_glPushMatrix,
			"function(:void)", OPT_SIDE_EFFECT);
  add_function_constant("glPushName", f_glPushName,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glRasterPos", f_glRasterPos,
			"function(float|int|array(float|int),float|int|void,float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glReadBuffer", f_glReadBuffer,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glRenderMode", f_glRenderMode,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glRotate", f_glRotate,
			"function(float|int|array(float|int),float|int|void,float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glScale", f_glScale,
			"function(float|int|array(float|int),float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glScissor", f_glScissor,
			"function(int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glShadeModel", f_glShadeModel,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glStencilFunc", f_glStencilFunc,
			"function(int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glStencilMask", f_glStencilMask,
			"function(int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glStencilOp", f_glStencilOp,
			"function(int,int,int:void)", OPT_SIDE_EFFECT);
  add_function_constant("glTexCoord", f_glTexCoord,
			"function(float|int|array(float|int),float|int|void,float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glTexEnv", f_glTexEnv,
			"function(int,int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glTexGen", f_glTexGen,
			"function(int,int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glTexImage2D", f_glTexImage2D,
			"function(int,int,int,int,object|mapping(string:object):void)", OPT_SIDE_EFFECT);
  add_function_constant("glTexParameter", f_glTexParameter,
			"function(int,int,float|int|array(float|int):void)", OPT_SIDE_EFFECT);
  add_function_constant("glTexSubImage2D", f_glTexSubImage2D,
			"function(int,int,int,int,object|mapping(string:object):void)", OPT_SIDE_EFFECT);
  add_function_constant("glTranslate", f_glTranslate,
			"function(float|int|array(float|int),float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glVertex", f_glVertex,
			"function(float|int|array(float|int),float|int|void,float|int|void,float|int|void:void)", OPT_SIDE_EFFECT);
  add_function_constant("glViewport", f_glViewport,
			"function(int,int,int,int:void)", OPT_SIDE_EFFECT);
  add_integer_constant("GL_2D", 1536, 0);
  add_integer_constant("GL_2_BYTES", 5127, 0);
  add_integer_constant("GL_3D", 1537, 0);
  add_integer_constant("GL_3D_COLOR", 1538, 0);
  add_integer_constant("GL_3D_COLOR_TEXTURE", 1539, 0);
  add_integer_constant("GL_3_BYTES", 5128, 0);
  add_integer_constant("GL_4D_COLOR_TEXTURE", 1540, 0);
  add_integer_constant("GL_4_BYTES", 5129, 0);
  add_integer_constant("GL_ABGR_EXT", 32768, 0);
  add_integer_constant("GL_ACCUM", 256, 0);
  add_integer_constant("GL_ACCUM_ALPHA_BITS", 3419, 0);
  add_integer_constant("GL_ACCUM_BLUE_BITS", 3418, 0);
  add_integer_constant("GL_ACCUM_BUFFER_BIT", 512, 0);
  add_integer_constant("GL_ACCUM_CLEAR_VALUE", 2944, 0);
  add_integer_constant("GL_ACCUM_GREEN_BITS", 3417, 0);
  add_integer_constant("GL_ACCUM_RED_BITS", 3416, 0);
  add_integer_constant("GL_ADD", 260, 0);
  add_integer_constant("GL_ALL_ATTRIB_BITS", 1048575, 0);
  add_integer_constant("GL_ALPHA", 6406, 0);
  add_integer_constant("GL_ALPHA12", 32829, 0);
  add_integer_constant("GL_ALPHA16", 32830, 0);
  add_integer_constant("GL_ALPHA4", 32827, 0);
  add_integer_constant("GL_ALPHA8", 32828, 0);
  add_integer_constant("GL_ALPHA_BIAS", 3357, 0);
  add_integer_constant("GL_ALPHA_BITS", 3413, 0);
  add_integer_constant("GL_ALPHA_SCALE", 3356, 0);
  add_integer_constant("GL_ALPHA_TEST", 3008, 0);
  add_integer_constant("GL_ALPHA_TEST_FUNC", 3009, 0);
  add_integer_constant("GL_ALPHA_TEST_REF", 3010, 0);
  add_integer_constant("GL_ALWAYS", 519, 0);
  add_integer_constant("GL_AMBIENT", 4608, 0);
  add_integer_constant("GL_AMBIENT_AND_DIFFUSE", 5634, 0);
  add_integer_constant("GL_AND", 5377, 0);
  add_integer_constant("GL_AND_INVERTED", 5380, 0);
  add_integer_constant("GL_AND_REVERSE", 5378, 0);
  add_integer_constant("GL_ATTRIB_STACK_DEPTH", 2992, 0);
  add_integer_constant("GL_AUTO_NORMAL", 3456, 0);
  add_integer_constant("GL_AUX0", 1033, 0);
  add_integer_constant("GL_AUX1", 1034, 0);
  add_integer_constant("GL_AUX2", 1035, 0);
  add_integer_constant("GL_AUX3", 1036, 0);
  add_integer_constant("GL_AUX_BUFFERS", 3072, 0);
  add_integer_constant("GL_AVERAGE_EXT", 33589, 0);
  add_integer_constant("GL_BACK", 1029, 0);
  add_integer_constant("GL_BACK_LEFT", 1026, 0);
  add_integer_constant("GL_BACK_RIGHT", 1027, 0);
  add_integer_constant("GL_BITMAP", 6656, 0);
  add_integer_constant("GL_BITMAP_TOKEN", 1796, 0);
  add_integer_constant("GL_BLEND", 3042, 0);
  add_integer_constant("GL_BLEND_COLOR_EXT", 32773, 0);
  add_integer_constant("GL_BLEND_DST", 3040, 0);
  add_integer_constant("GL_BLEND_EQUATION_EXT", 32777, 0);
  add_integer_constant("GL_BLEND_SRC", 3041, 0);
  add_integer_constant("GL_BLUE", 6405, 0);
  add_integer_constant("GL_BLUE_BIAS", 3355, 0);
  add_integer_constant("GL_BLUE_BITS", 3412, 0);
  add_integer_constant("GL_BLUE_SCALE", 3354, 0);
  add_integer_constant("GL_BYTE", 5120, 0);
  add_integer_constant("GL_C3F_V3F", 10788, 0);
  add_integer_constant("GL_C4F_N3F_V3F", 10790, 0);
  add_integer_constant("GL_C4UB_V2F", 10786, 0);
  add_integer_constant("GL_C4UB_V3F", 10787, 0);
  add_integer_constant("GL_CCW", 2305, 0);
  add_integer_constant("GL_CLAMP", 10496, 0);
  add_integer_constant("GL_CLEAR", 5376, 0);
  add_integer_constant("GL_CLIENT_ALL_ATTRIB_BITS", 4294967295, 0);
  add_integer_constant("GL_CLIENT_ATTRIB_STACK_DEPTH", 2993, 0);
  add_integer_constant("GL_CLIENT_PIXEL_STORE_BIT", 1, 0);
  add_integer_constant("GL_CLIENT_VERTEX_ARRAY_BIT", 2, 0);
  add_integer_constant("GL_CLIP_PLANE0", 12288, 0);
  add_integer_constant("GL_CLIP_PLANE1", 12289, 0);
  add_integer_constant("GL_CLIP_PLANE2", 12290, 0);
  add_integer_constant("GL_CLIP_PLANE3", 12291, 0);
  add_integer_constant("GL_CLIP_PLANE4", 12292, 0);
  add_integer_constant("GL_CLIP_PLANE5", 12293, 0);
  add_integer_constant("GL_COEFF", 2560, 0);
  add_integer_constant("GL_COLOR", 6144, 0);
  add_integer_constant("GL_COLOR_ARRAY", 32886, 0);
  add_integer_constant("GL_COLOR_ARRAY_POINTER", 32912, 0);
  add_integer_constant("GL_COLOR_ARRAY_SIZE", 32897, 0);
  add_integer_constant("GL_COLOR_ARRAY_STRIDE", 32899, 0);
  add_integer_constant("GL_COLOR_ARRAY_TYPE", 32898, 0);
  add_integer_constant("GL_COLOR_BUFFER_BIT", 16384, 0);
  add_integer_constant("GL_COLOR_CLEAR_VALUE", 3106, 0);
  add_integer_constant("GL_COLOR_INDEX", 6400, 0);
  add_integer_constant("GL_COLOR_INDEXES", 5635, 0);
  add_integer_constant("GL_COLOR_LOGIC_OP", 3058, 0);
  add_integer_constant("GL_COLOR_MATERIAL", 2903, 0);
  add_integer_constant("GL_COLOR_MATERIAL_FACE", 2901, 0);
  add_integer_constant("GL_COLOR_MATERIAL_PARAMETER", 2902, 0);
  add_integer_constant("GL_COLOR_TABLE_ALPHA_SIZE_SGI", 32989, 0);
  add_integer_constant("GL_COLOR_TABLE_BIAS_SGI", 32983, 0);
  add_integer_constant("GL_COLOR_TABLE_BLUE_SIZE_SGI", 32988, 0);
  add_integer_constant("GL_COLOR_TABLE_FORMAT_SGI", 32984, 0);
  add_integer_constant("GL_COLOR_TABLE_GREEN_SIZE_SGI", 32987, 0);
  add_integer_constant("GL_COLOR_TABLE_INTENSITY_SIZE_SGI", 32991, 0);
  add_integer_constant("GL_COLOR_TABLE_LUMINANCE_SIZE_SGI", 32990, 0);
  add_integer_constant("GL_COLOR_TABLE_RED_SIZE_SGI", 32986, 0);
  add_integer_constant("GL_COLOR_TABLE_SCALE_SGI", 32982, 0);
  add_integer_constant("GL_COLOR_TABLE_SGI", 32976, 0);
  add_integer_constant("GL_COLOR_TABLE_WIDTH_SGI", 32985, 0);
  add_integer_constant("GL_COLOR_WRITEMASK", 3107, 0);
  add_integer_constant("GL_COMPILE", 4864, 0);
  add_integer_constant("GL_COMPILE_AND_EXECUTE", 4865, 0);
  add_integer_constant("GL_COMPRESSED_GEOM_ACCELERATED_SUNX", 33232, 0);
  add_integer_constant("GL_COMPRESSED_GEOM_VERSION_SUNX", 33233, 0);
  add_integer_constant("GL_CONSTANT_ALPHA_EXT", 32771, 0);
  add_integer_constant("GL_CONSTANT_ATTENUATION", 4615, 0);
  add_integer_constant("GL_CONSTANT_BORDER_HP", 33105, 0);
  add_integer_constant("GL_CONSTANT_COLOR_EXT", 32769, 0);
  add_integer_constant("GL_CONVOLUTION_1D_EXT", 32784, 0);
  add_integer_constant("GL_CONVOLUTION_2D_EXT", 32785, 0);
  add_integer_constant("GL_CONVOLUTION_BORDER_COLOR_HP", 33108, 0);
  add_integer_constant("GL_CONVOLUTION_BORDER_MODE_EXT", 32787, 0);
  add_integer_constant("GL_CONVOLUTION_FILTER_BIAS_EXT", 32789, 0);
  add_integer_constant("GL_CONVOLUTION_FILTER_SCALE_EXT", 32788, 0);
  add_integer_constant("GL_CONVOLUTION_FORMAT_EXT", 32791, 0);
  add_integer_constant("GL_CONVOLUTION_HEIGHT_EXT", 32793, 0);
  add_integer_constant("GL_CONVOLUTION_WIDTH_EXT", 32792, 0);
  add_integer_constant("GL_COPY", 5379, 0);
  add_integer_constant("GL_COPY_INVERTED", 5388, 0);
  add_integer_constant("GL_COPY_PIXEL_TOKEN", 1798, 0);
  add_integer_constant("GL_CUBIC_EXT", 33588, 0);
  add_integer_constant("GL_CULL_FACE", 2884, 0);
  add_integer_constant("GL_CULL_FACE_MODE", 2885, 0);
  add_integer_constant("GL_CURRENT_BIT", 1, 0);
  add_integer_constant("GL_CURRENT_COLOR", 2816, 0);
  add_integer_constant("GL_CURRENT_INDEX", 2817, 0);
  add_integer_constant("GL_CURRENT_NORMAL", 2818, 0);
  add_integer_constant("GL_CURRENT_RASTER_COLOR", 2820, 0);
  add_integer_constant("GL_CURRENT_RASTER_DISTANCE", 2825, 0);
  add_integer_constant("GL_CURRENT_RASTER_INDEX", 2821, 0);
  add_integer_constant("GL_CURRENT_RASTER_POSITION", 2823, 0);
  add_integer_constant("GL_CURRENT_RASTER_POSITION_VALID", 2824, 0);
  add_integer_constant("GL_CURRENT_RASTER_TEXTURE_COORDS", 2822, 0);
  add_integer_constant("GL_CURRENT_TEXTURE_COORDS", 2819, 0);
  add_integer_constant("GL_CW", 2304, 0);
  add_integer_constant("GL_DECAL", 8449, 0);
  add_integer_constant("GL_DECR", 7683, 0);
  add_integer_constant("GL_DEPTH", 6145, 0);
  add_integer_constant("GL_DEPTH_BIAS", 3359, 0);
  add_integer_constant("GL_DEPTH_BITS", 3414, 0);
  add_integer_constant("GL_DEPTH_BUFFER_BIT", 256, 0);
  add_integer_constant("GL_DEPTH_CLEAR_VALUE", 2931, 0);
  add_integer_constant("GL_DEPTH_COMPONENT", 6402, 0);
  add_integer_constant("GL_DEPTH_FUNC", 2932, 0);
  add_integer_constant("GL_DEPTH_RANGE", 2928, 0);
  add_integer_constant("GL_DEPTH_SCALE", 3358, 0);
  add_integer_constant("GL_DEPTH_TEST", 2929, 0);
  add_integer_constant("GL_DEPTH_WRITEMASK", 2930, 0);
  add_integer_constant("GL_DIFFUSE", 4609, 0);
  add_integer_constant("GL_DITHER", 3024, 0);
  add_integer_constant("GL_DOMAIN", 2562, 0);
  add_integer_constant("GL_DONT_CARE", 4352, 0);
  add_integer_constant("GL_DOUBLE", 5130, 0);
  add_integer_constant("GL_DOUBLEBUFFER", 3122, 0);
  add_integer_constant("GL_DRAW_BUFFER", 3073, 0);
  add_integer_constant("GL_DRAW_PIXEL_TOKEN", 1797, 0);
  add_integer_constant("GL_DST_ALPHA", 772, 0);
  add_integer_constant("GL_DST_COLOR", 774, 0);
  add_integer_constant("GL_EDGE_FLAG", 2883, 0);
  add_integer_constant("GL_EDGE_FLAG_ARRAY", 32889, 0);
  add_integer_constant("GL_EDGE_FLAG_ARRAY_POINTER", 32915, 0);
  add_integer_constant("GL_EDGE_FLAG_ARRAY_STRIDE", 32908, 0);
  add_integer_constant("GL_EMISSION", 5632, 0);
  add_integer_constant("GL_ENABLE_BIT", 8192, 0);
  add_integer_constant("GL_EQUAL", 514, 0);
  add_integer_constant("GL_EQUIV", 5385, 0);
  add_integer_constant("GL_EVAL_BIT", 65536, 0);
  add_integer_constant("GL_EXP", 2048, 0);
  add_integer_constant("GL_EXP2", 2049, 0);
  add_integer_constant("GL_EXTENSIONS", 7939, 0);
  add_integer_constant("GL_EXT_abgr", 1, 0);
  add_integer_constant("GL_EXT_blend_color", 1, 0);
  add_integer_constant("GL_EXT_blend_minmax", 1, 0);
  add_integer_constant("GL_EXT_blend_subtract", 1, 0);
  add_integer_constant("GL_EXT_convolution", 1, 0);
  add_integer_constant("GL_EXT_histogram", 1, 0);
  add_integer_constant("GL_EXT_pixel_transform", 1, 0);
  add_integer_constant("GL_EXT_rescale_normal", 1, 0);
  add_integer_constant("GL_EXT_texture3D", 1, 0);
  add_integer_constant("GL_EYE_LINEAR", 9216, 0);
  add_integer_constant("GL_EYE_PLANE", 9474, 0);
  add_integer_constant("GL_FALSE", 0, 0);
  add_integer_constant("GL_FASTEST", 4353, 0);
  add_integer_constant("GL_FEEDBACK", 7169, 0);
  add_integer_constant("GL_FEEDBACK_BUFFER_POINTER", 3568, 0);
  add_integer_constant("GL_FEEDBACK_BUFFER_SIZE", 3569, 0);
  add_integer_constant("GL_FEEDBACK_BUFFER_TYPE", 3570, 0);
  add_integer_constant("GL_FILL", 6914, 0);
  add_integer_constant("GL_FLAT", 7424, 0);
  add_integer_constant("GL_FLOAT", 5126, 0);
  add_integer_constant("GL_FOG", 2912, 0);
  add_integer_constant("GL_FOG_BIT", 128, 0);
  add_integer_constant("GL_FOG_COLOR", 2918, 0);
  add_integer_constant("GL_FOG_DENSITY", 2914, 0);
  add_integer_constant("GL_FOG_END", 2916, 0);
  add_integer_constant("GL_FOG_HINT", 3156, 0);
  add_integer_constant("GL_FOG_INDEX", 2913, 0);
  add_integer_constant("GL_FOG_MODE", 2917, 0);
  add_integer_constant("GL_FOG_START", 2915, 0);
  add_integer_constant("GL_FRONT", 1028, 0);
  add_integer_constant("GL_FRONT_AND_BACK", 1032, 0);
  add_integer_constant("GL_FRONT_FACE", 2886, 0);
  add_integer_constant("GL_FRONT_LEFT", 1024, 0);
  add_integer_constant("GL_FRONT_RIGHT", 1025, 0);
  add_integer_constant("GL_FUNC_ADD_EXT", 32774, 0);
  add_integer_constant("GL_FUNC_REVERSE_SUBTRACT_EXT", 32779, 0);
  add_integer_constant("GL_FUNC_SUBTRACT_EXT", 32778, 0);
  add_integer_constant("GL_GEQUAL", 518, 0);
  add_integer_constant("GL_GREATER", 516, 0);
  add_integer_constant("GL_GREEN", 6404, 0);
  add_integer_constant("GL_GREEN_BIAS", 3353, 0);
  add_integer_constant("GL_GREEN_BITS", 3411, 0);
  add_integer_constant("GL_GREEN_SCALE", 3352, 0);
  add_integer_constant("GL_HINT_BIT", 32768, 0);
  add_integer_constant("GL_HISTOGRAM_ALPHA_SIZE_EXT", 32811, 0);
  add_integer_constant("GL_HISTOGRAM_BLUE_SIZE_EXT", 32810, 0);
  add_integer_constant("GL_HISTOGRAM_EXT", 32804, 0);
  add_integer_constant("GL_HISTOGRAM_FORMAT_EXT", 32807, 0);
  add_integer_constant("GL_HISTOGRAM_GREEN_SIZE_EXT", 32809, 0);
  add_integer_constant("GL_HISTOGRAM_LUMINANCE_SIZE_EXT", 32812, 0);
  add_integer_constant("GL_HISTOGRAM_RED_SIZE_EXT", 32808, 0);
  add_integer_constant("GL_HISTOGRAM_SINK_EXT", 32813, 0);
  add_integer_constant("GL_HISTOGRAM_WIDTH_EXT", 32806, 0);
  add_integer_constant("GL_HP_convolution_border_modes", 1, 0);
  add_integer_constant("GL_HP_occlusion_test", 1, 0);
  add_integer_constant("GL_IGNORE_BORDER_HP", 33104, 0);
  add_integer_constant("GL_INCR", 7682, 0);
  add_integer_constant("GL_INDEX_ARRAY", 32887, 0);
  add_integer_constant("GL_INDEX_ARRAY_POINTER", 32913, 0);
  add_integer_constant("GL_INDEX_ARRAY_STRIDE", 32902, 0);
  add_integer_constant("GL_INDEX_ARRAY_TYPE", 32901, 0);
  add_integer_constant("GL_INDEX_BITS", 3409, 0);
  add_integer_constant("GL_INDEX_CLEAR_VALUE", 3104, 0);
  add_integer_constant("GL_INDEX_LOGIC_OP", 3057, 0);
  add_integer_constant("GL_INDEX_MODE", 3120, 0);
  add_integer_constant("GL_INDEX_OFFSET", 3347, 0);
  add_integer_constant("GL_INDEX_SHIFT", 3346, 0);
  add_integer_constant("GL_INDEX_WRITEMASK", 3105, 0);
  add_integer_constant("GL_INT", 5124, 0);
  add_integer_constant("GL_INTENSITY", 32841, 0);
  add_integer_constant("GL_INTENSITY12", 32844, 0);
  add_integer_constant("GL_INTENSITY16", 32845, 0);
  add_integer_constant("GL_INTENSITY4", 32842, 0);
  add_integer_constant("GL_INTENSITY8", 32843, 0);
  add_integer_constant("GL_INVALID_ENUM", 1280, 0);
  add_integer_constant("GL_INVALID_OPERATION", 1282, 0);
  add_integer_constant("GL_INVALID_VALUE", 1281, 0);
  add_integer_constant("GL_INVERT", 5386, 0);
  add_integer_constant("GL_KEEP", 7680, 0);
  add_integer_constant("GL_LARGE_SUNX", 33235, 0);
  add_integer_constant("GL_LEFT", 1030, 0);
  add_integer_constant("GL_LEQUAL", 515, 0);
  add_integer_constant("GL_LESS", 513, 0);
  add_integer_constant("GL_LIGHT0", 16384, 0);
  add_integer_constant("GL_LIGHT1", 16385, 0);
  add_integer_constant("GL_LIGHT2", 16386, 0);
  add_integer_constant("GL_LIGHT3", 16387, 0);
  add_integer_constant("GL_LIGHT4", 16388, 0);
  add_integer_constant("GL_LIGHT5", 16389, 0);
  add_integer_constant("GL_LIGHT6", 16390, 0);
  add_integer_constant("GL_LIGHT7", 16391, 0);
  add_integer_constant("GL_LIGHTING", 2896, 0);
  add_integer_constant("GL_LIGHTING_BIT", 64, 0);
  add_integer_constant("GL_LIGHT_MODEL_AMBIENT", 2899, 0);
  add_integer_constant("GL_LIGHT_MODEL_LOCAL_VIEWER", 2897, 0);
  add_integer_constant("GL_LIGHT_MODEL_TWO_SIDE", 2898, 0);
  add_integer_constant("GL_LINE", 6913, 0);
  add_integer_constant("GL_LINEAR", 9729, 0);
  add_integer_constant("GL_LINEAR_ATTENUATION", 4616, 0);
  add_integer_constant("GL_LINEAR_MIPMAP_LINEAR", 9987, 0);
  add_integer_constant("GL_LINEAR_MIPMAP_NEAREST", 9985, 0);
  add_integer_constant("GL_LINES", 1, 0);
  add_integer_constant("GL_LINE_BIT", 4, 0);
  add_integer_constant("GL_LINE_LOOP", 2, 0);
  add_integer_constant("GL_LINE_RESET_TOKEN", 1799, 0);
  add_integer_constant("GL_LINE_SMOOTH", 2848, 0);
  add_integer_constant("GL_LINE_SMOOTH_HINT", 3154, 0);
  add_integer_constant("GL_LINE_STIPPLE", 2852, 0);
  add_integer_constant("GL_LINE_STIPPLE_PATTERN", 2853, 0);
  add_integer_constant("GL_LINE_STIPPLE_REPEAT", 2854, 0);
  add_integer_constant("GL_LINE_STRIP", 3, 0);
  add_integer_constant("GL_LINE_TOKEN", 1794, 0);
  add_integer_constant("GL_LINE_WIDTH", 2849, 0);
  add_integer_constant("GL_LINE_WIDTH_GRANULARITY", 2851, 0);
  add_integer_constant("GL_LINE_WIDTH_RANGE", 2850, 0);
  add_integer_constant("GL_LIST_BASE", 2866, 0);
  add_integer_constant("GL_LIST_BIT", 131072, 0);
  add_integer_constant("GL_LIST_INDEX", 2867, 0);
  add_integer_constant("GL_LIST_MODE", 2864, 0);
  add_integer_constant("GL_LOAD", 257, 0);
  add_integer_constant("GL_LOGIC_OP", 3057, 0);
  add_integer_constant("GL_LOGIC_OP_MODE", 3056, 0);
  add_integer_constant("GL_LUMINANCE", 6409, 0);
  add_integer_constant("GL_LUMINANCE12", 32833, 0);
  add_integer_constant("GL_LUMINANCE12_ALPHA12", 32839, 0);
  add_integer_constant("GL_LUMINANCE12_ALPHA4", 32838, 0);
  add_integer_constant("GL_LUMINANCE16", 32834, 0);
  add_integer_constant("GL_LUMINANCE16_ALPHA16", 32840, 0);
  add_integer_constant("GL_LUMINANCE4", 32831, 0);
  add_integer_constant("GL_LUMINANCE4_ALPHA4", 32835, 0);
  add_integer_constant("GL_LUMINANCE6_ALPHA2", 32836, 0);
  add_integer_constant("GL_LUMINANCE8", 32832, 0);
  add_integer_constant("GL_LUMINANCE8_ALPHA8", 32837, 0);
  add_integer_constant("GL_LUMINANCE_ALPHA", 6410, 0);
  add_integer_constant("GL_MAP1_COLOR_4", 3472, 0);
  add_integer_constant("GL_MAP1_GRID_DOMAIN", 3536, 0);
  add_integer_constant("GL_MAP1_GRID_SEGMENTS", 3537, 0);
  add_integer_constant("GL_MAP1_INDEX", 3473, 0);
  add_integer_constant("GL_MAP1_NORMAL", 3474, 0);
  add_integer_constant("GL_MAP1_TEXTURE_COORD_1", 3475, 0);
  add_integer_constant("GL_MAP1_TEXTURE_COORD_2", 3476, 0);
  add_integer_constant("GL_MAP1_TEXTURE_COORD_3", 3477, 0);
  add_integer_constant("GL_MAP1_TEXTURE_COORD_4", 3478, 0);
  add_integer_constant("GL_MAP1_VERTEX_3", 3479, 0);
  add_integer_constant("GL_MAP1_VERTEX_4", 3480, 0);
  add_integer_constant("GL_MAP2_COLOR_4", 3504, 0);
  add_integer_constant("GL_MAP2_GRID_DOMAIN", 3538, 0);
  add_integer_constant("GL_MAP2_GRID_SEGMENTS", 3539, 0);
  add_integer_constant("GL_MAP2_INDEX", 3505, 0);
  add_integer_constant("GL_MAP2_NORMAL", 3506, 0);
  add_integer_constant("GL_MAP2_TEXTURE_COORD_1", 3507, 0);
  add_integer_constant("GL_MAP2_TEXTURE_COORD_2", 3508, 0);
  add_integer_constant("GL_MAP2_TEXTURE_COORD_3", 3509, 0);
  add_integer_constant("GL_MAP2_TEXTURE_COORD_4", 3510, 0);
  add_integer_constant("GL_MAP2_VERTEX_3", 3511, 0);
  add_integer_constant("GL_MAP2_VERTEX_4", 3512, 0);
  add_integer_constant("GL_MAP_COLOR", 3344, 0);
  add_integer_constant("GL_MAP_STENCIL", 3345, 0);
  add_integer_constant("GL_MATRIX_MODE", 2976, 0);
  add_integer_constant("GL_MAX_3D_TEXTURE_SIZE_EXT", 32883, 0);
  add_integer_constant("GL_MAX_ATTRIB_STACK_DEPTH", 3381, 0);
  add_integer_constant("GL_MAX_CLIENT_ATTRIB_STACK_DEPTH", 3387, 0);
  add_integer_constant("GL_MAX_CLIP_PLANES", 3378, 0);
  add_integer_constant("GL_MAX_CONVOLUTION_HEIGHT_EXT", 32795, 0);
  add_integer_constant("GL_MAX_CONVOLUTION_WIDTH_EXT", 32794, 0);
  add_integer_constant("GL_MAX_EVAL_ORDER", 3376, 0);
  add_integer_constant("GL_MAX_EXT", 32776, 0);
  add_integer_constant("GL_MAX_LIGHTS", 3377, 0);
  add_integer_constant("GL_MAX_LIST_NESTING", 2865, 0);
  add_integer_constant("GL_MAX_MODELVIEW_STACK_DEPTH", 3382, 0);
  add_integer_constant("GL_MAX_NAME_STACK_DEPTH", 3383, 0);
  add_integer_constant("GL_MAX_PIXEL_MAP_TABLE", 3380, 0);
  add_integer_constant("GL_MAX_PIXEL_TRANSFORM_2D_STACK_DEPTH_EXT", 33591, 0);
  add_integer_constant("GL_MAX_PROJECTION_STACK_DEPTH", 3384, 0);
  add_integer_constant("GL_MAX_TEXTURE_SIZE", 3379, 0);
  add_integer_constant("GL_MAX_TEXTURE_STACK_DEPTH", 3385, 0);
  add_integer_constant("GL_MAX_VIEWPORT_DIMS", 3386, 0);
  add_integer_constant("GL_MINMAX_EXT", 32814, 0);
  add_integer_constant("GL_MINMAX_FORMAT_EXT", 32815, 0);
  add_integer_constant("GL_MINMAX_SINK_EXT", 32816, 0);
  add_integer_constant("GL_MIN_EXT", 32775, 0);
  add_integer_constant("GL_MODELVIEW", 5888, 0);
  add_integer_constant("GL_MODELVIEW_MATRIX", 2982, 0);
  add_integer_constant("GL_MODELVIEW_STACK_DEPTH", 2979, 0);
  add_integer_constant("GL_MODULATE", 8448, 0);
  add_integer_constant("GL_MULT", 259, 0);
  add_integer_constant("GL_N3F_V3F", 10789, 0);
  add_integer_constant("GL_NAME_STACK_DEPTH", 3440, 0);
  add_integer_constant("GL_NAND", 5390, 0);
  add_integer_constant("GL_NEAREST", 9728, 0);
  add_integer_constant("GL_NEAREST_MIPMAP_LINEAR", 9986, 0);
  add_integer_constant("GL_NEAREST_MIPMAP_NEAREST", 9984, 0);
  add_integer_constant("GL_NEVER", 512, 0);
  add_integer_constant("GL_NICEST", 4354, 0);
  add_integer_constant("GL_NONE", 0, 0);
  add_integer_constant("GL_NOOP", 5381, 0);
  add_integer_constant("GL_NOR", 5384, 0);
  add_integer_constant("GL_NORMALIZE", 2977, 0);
  add_integer_constant("GL_NORMAL_ARRAY", 32885, 0);
  add_integer_constant("GL_NORMAL_ARRAY_POINTER", 32911, 0);
  add_integer_constant("GL_NORMAL_ARRAY_STRIDE", 32895, 0);
  add_integer_constant("GL_NORMAL_ARRAY_TYPE", 32894, 0);
  add_integer_constant("GL_NOTEQUAL", 517, 0);
  add_integer_constant("GL_NO_ERROR", 0, 0);
  add_integer_constant("GL_OBJECT_LINEAR", 9217, 0);
  add_integer_constant("GL_OBJECT_PLANE", 9473, 0);
  add_integer_constant("GL_OCCLUSION_RESULT_HP", 33126, 0);
  add_integer_constant("GL_OCCLUSION_TEST_HP", 33125, 0);
  add_integer_constant("GL_ONE", 1, 0);
  add_integer_constant("GL_ONE_MINUS_CONSTANT_ALPHA_EXT", 32772, 0);
  add_integer_constant("GL_ONE_MINUS_CONSTANT_COLOR_EXT", 32770, 0);
  add_integer_constant("GL_ONE_MINUS_DST_ALPHA", 773, 0);
  add_integer_constant("GL_ONE_MINUS_DST_COLOR", 775, 0);
  add_integer_constant("GL_ONE_MINUS_SRC_ALPHA", 771, 0);
  add_integer_constant("GL_ONE_MINUS_SRC_COLOR", 769, 0);
  add_integer_constant("GL_OR", 5383, 0);
  add_integer_constant("GL_ORDER", 2561, 0);
  add_integer_constant("GL_OR_INVERTED", 5389, 0);
  add_integer_constant("GL_OR_REVERSE", 5387, 0);
  add_integer_constant("GL_OUT_OF_MEMORY", 1285, 0);
  add_integer_constant("GL_PACK_ALIGNMENT", 3333, 0);
  add_integer_constant("GL_PACK_IMAGE_HEIGHT_EXT", 32876, 0);
  add_integer_constant("GL_PACK_LSB_FIRST", 3329, 0);
  add_integer_constant("GL_PACK_ROW_LENGTH", 3330, 0);
  add_integer_constant("GL_PACK_SKIP_IMAGES_EXT", 32875, 0);
  add_integer_constant("GL_PACK_SKIP_PIXELS", 3332, 0);
  add_integer_constant("GL_PACK_SKIP_ROWS", 3331, 0);
  add_integer_constant("GL_PACK_SWAP_BYTES", 3328, 0);
  add_integer_constant("GL_PASS_THROUGH_TOKEN", 1792, 0);
  add_integer_constant("GL_PERSPECTIVE_CORRECTION_HINT", 3152, 0);
  add_integer_constant("GL_PIXEL_CUBIC_WEIGHT_EXT", 33587, 0);
  add_integer_constant("GL_PIXEL_MAG_FILTER_EXT", 33585, 0);
  add_integer_constant("GL_PIXEL_MAP_A_TO_A", 3193, 0);
  add_integer_constant("GL_PIXEL_MAP_A_TO_A_SIZE", 3257, 0);
  add_integer_constant("GL_PIXEL_MAP_B_TO_B", 3192, 0);
  add_integer_constant("GL_PIXEL_MAP_B_TO_B_SIZE", 3256, 0);
  add_integer_constant("GL_PIXEL_MAP_G_TO_G", 3191, 0);
  add_integer_constant("GL_PIXEL_MAP_G_TO_G_SIZE", 3255, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_A", 3189, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_A_SIZE", 3253, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_B", 3188, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_B_SIZE", 3252, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_G", 3187, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_G_SIZE", 3251, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_I", 3184, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_I_SIZE", 3248, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_R", 3186, 0);
  add_integer_constant("GL_PIXEL_MAP_I_TO_R_SIZE", 3250, 0);
  add_integer_constant("GL_PIXEL_MAP_R_TO_R", 3190, 0);
  add_integer_constant("GL_PIXEL_MAP_R_TO_R_SIZE", 3254, 0);
  add_integer_constant("GL_PIXEL_MAP_S_TO_S", 3185, 0);
  add_integer_constant("GL_PIXEL_MAP_S_TO_S_SIZE", 3249, 0);
  add_integer_constant("GL_PIXEL_MIN_FILTER_EXT", 33586, 0);
  add_integer_constant("GL_PIXEL_MODE_BIT", 32, 0);
  add_integer_constant("GL_PIXEL_TRANSFORM_2D_EXT", 33584, 0);
  add_integer_constant("GL_PIXEL_TRANSFORM_2D_MATRIX_EXT", 33592, 0);
  add_integer_constant("GL_PIXEL_TRANSFORM_2D_STACK_DEPTH_EXT", 33590, 0);
  add_integer_constant("GL_PIXEL_TRANSFORM_COLOR_TABLE_EXT", 33593, 0);
  add_integer_constant("GL_POINT", 6912, 0);
  add_integer_constant("GL_POINTS", 0, 0);
  add_integer_constant("GL_POINT_BIT", 2, 0);
  add_integer_constant("GL_POINT_SIZE", 2833, 0);
  add_integer_constant("GL_POINT_SIZE_GRANULARITY", 2835, 0);
  add_integer_constant("GL_POINT_SIZE_RANGE", 2834, 0);
  add_integer_constant("GL_POINT_SMOOTH", 2832, 0);
  add_integer_constant("GL_POINT_SMOOTH_HINT", 3153, 0);
  add_integer_constant("GL_POINT_TOKEN", 1793, 0);
  add_integer_constant("GL_POLYGON", 9, 0);
  add_integer_constant("GL_POLYGON_BIT", 8, 0);
  add_integer_constant("GL_POLYGON_MODE", 2880, 0);
  add_integer_constant("GL_POLYGON_OFFSET_FACTOR", 32824, 0);
  add_integer_constant("GL_POLYGON_OFFSET_FILL", 32823, 0);
  add_integer_constant("GL_POLYGON_OFFSET_LINE", 10754, 0);
  add_integer_constant("GL_POLYGON_OFFSET_POINT", 10753, 0);
  add_integer_constant("GL_POLYGON_OFFSET_UNITS", 10752, 0);
  add_integer_constant("GL_POLYGON_SMOOTH", 2881, 0);
  add_integer_constant("GL_POLYGON_SMOOTH_HINT", 3155, 0);
  add_integer_constant("GL_POLYGON_STIPPLE", 2882, 0);
  add_integer_constant("GL_POLYGON_STIPPLE_BIT", 16, 0);
  add_integer_constant("GL_POLYGON_TOKEN", 1795, 0);
  add_integer_constant("GL_POSITION", 4611, 0);
  add_integer_constant("GL_POST_CONVOLUTION_ALPHA_BIAS_EXT", 32803, 0);
  add_integer_constant("GL_POST_CONVOLUTION_ALPHA_SCALE_EXT", 32799, 0);
  add_integer_constant("GL_POST_CONVOLUTION_BLUE_BIAS_EXT", 32802, 0);
  add_integer_constant("GL_POST_CONVOLUTION_BLUE_SCALE_EXT", 32798, 0);
  add_integer_constant("GL_POST_CONVOLUTION_COLOR_TABLE_SGI", 32977, 0);
  add_integer_constant("GL_POST_CONVOLUTION_GREEN_BIAS_EXT", 32801, 0);
  add_integer_constant("GL_POST_CONVOLUTION_GREEN_SCALE_EXT", 32797, 0);
  add_integer_constant("GL_POST_CONVOLUTION_RED_BIAS_EXT", 32800, 0);
  add_integer_constant("GL_POST_CONVOLUTION_RED_SCALE_EXT", 32796, 0);
  add_integer_constant("GL_PROJECTION", 5889, 0);
  add_integer_constant("GL_PROJECTION_MATRIX", 2983, 0);
  add_integer_constant("GL_PROJECTION_STACK_DEPTH", 2980, 0);
  add_integer_constant("GL_PROXY_COLOR_TABLE_SGI", 32979, 0);
  add_integer_constant("GL_PROXY_HISTOGRAM_EXT", 32805, 0);
  add_integer_constant("GL_PROXY_PIXEL_TRANSFORM_COLOR_TABLE_EXT", 33594, 0);
  add_integer_constant("GL_PROXY_POST_CONVOLUTION_COLOR_TABLE_SGI", 32980, 0);
  add_integer_constant("GL_PROXY_TEXTURE_1D", 32867, 0);
  add_integer_constant("GL_PROXY_TEXTURE_2D", 32868, 0);
  add_integer_constant("GL_PROXY_TEXTURE_3D_EXT", 32880, 0);
  add_integer_constant("GL_PROXY_TEXTURE_COLOR_TABLE_SGI", 32957, 0);
  add_integer_constant("GL_Q", 8195, 0);
  add_integer_constant("GL_QUADRATIC_ATTENUATION", 4617, 0);
  add_integer_constant("GL_QUADS", 7, 0);
  add_integer_constant("GL_QUAD_STRIP", 8, 0);
  add_integer_constant("GL_R", 8194, 0);
  add_integer_constant("GL_R3_G3_B2", 10768, 0);
  add_integer_constant("GL_READ_BUFFER", 3074, 0);
  add_integer_constant("GL_RED", 6403, 0);
  add_integer_constant("GL_REDUCE_EXT", 32790, 0);
  add_integer_constant("GL_RED_BIAS", 3349, 0);
  add_integer_constant("GL_RED_BITS", 3410, 0);
  add_integer_constant("GL_RED_SCALE", 3348, 0);
  add_integer_constant("GL_RENDER", 7168, 0);
  add_integer_constant("GL_RENDERER", 7937, 0);
  add_integer_constant("GL_RENDER_MODE", 3136, 0);
  add_integer_constant("GL_REPEAT", 10497, 0);
  add_integer_constant("GL_REPLACE", 7681, 0);
  add_integer_constant("GL_REPLICATE_BORDER_HP", 33107, 0);
  add_integer_constant("GL_RESCALE_NORMAL_EXT", 32826, 0);
  add_integer_constant("GL_RETURN", 258, 0);
  add_integer_constant("GL_RGB", 6407, 0);
  add_integer_constant("GL_RGB10", 32850, 0);
  add_integer_constant("GL_RGB10_A2", 32857, 0);
  add_integer_constant("GL_RGB12", 32851, 0);
  add_integer_constant("GL_RGB16", 32852, 0);
  add_integer_constant("GL_RGB4", 32847, 0);
  add_integer_constant("GL_RGB5", 32848, 0);
  add_integer_constant("GL_RGB5_A1", 32855, 0);
  add_integer_constant("GL_RGB8", 32849, 0);
  add_integer_constant("GL_RGBA", 6408, 0);
  add_integer_constant("GL_RGBA12", 32858, 0);
  add_integer_constant("GL_RGBA16", 32859, 0);
  add_integer_constant("GL_RGBA2", 32853, 0);
  add_integer_constant("GL_RGBA4", 32854, 0);
  add_integer_constant("GL_RGBA8", 32856, 0);
  add_integer_constant("GL_RGBA_MODE", 3121, 0);
  add_integer_constant("GL_RIGHT", 1031, 0);
  add_integer_constant("GL_S", 8192, 0);
  add_integer_constant("GL_SCISSOR_BIT", 524288, 0);
  add_integer_constant("GL_SCISSOR_BOX", 3088, 0);
  add_integer_constant("GL_SCISSOR_TEST", 3089, 0);
  add_integer_constant("GL_SELECT", 7170, 0);
  add_integer_constant("GL_SELECTION_BUFFER_POINTER", 3571, 0);
  add_integer_constant("GL_SELECTION_BUFFER_SIZE", 3572, 0);
  add_integer_constant("GL_SEPARABLE_2D_EXT", 32786, 0);
  add_integer_constant("GL_SET", 5391, 0);
  add_integer_constant("GL_SGI_color_table", 1, 0);
  add_integer_constant("GL_SGI_texture_color_table", 1, 0);
  add_integer_constant("GL_SHADE_MODEL", 2900, 0);
  add_integer_constant("GL_SHININESS", 5633, 0);
  add_integer_constant("GL_SHORT", 5122, 0);
  add_integer_constant("GL_SMOOTH", 7425, 0);
  add_integer_constant("GL_SPECULAR", 4610, 0);
  add_integer_constant("GL_SPHERE_MAP", 9218, 0);
  add_integer_constant("GL_SPOT_CUTOFF", 4614, 0);
  add_integer_constant("GL_SPOT_DIRECTION", 4612, 0);
  add_integer_constant("GL_SPOT_EXPONENT", 4613, 0);
  add_integer_constant("GL_SRC_ALPHA", 770, 0);
  add_integer_constant("GL_SRC_ALPHA_SATURATE", 776, 0);
  add_integer_constant("GL_SRC_COLOR", 768, 0);
  add_integer_constant("GL_STACK_OVERFLOW", 1283, 0);
  add_integer_constant("GL_STACK_UNDERFLOW", 1284, 0);
  add_integer_constant("GL_STENCIL", 6146, 0);
  add_integer_constant("GL_STENCIL_BITS", 3415, 0);
  add_integer_constant("GL_STENCIL_BUFFER_BIT", 1024, 0);
  add_integer_constant("GL_STENCIL_CLEAR_VALUE", 2961, 0);
  add_integer_constant("GL_STENCIL_FAIL", 2964, 0);
  add_integer_constant("GL_STENCIL_FUNC", 2962, 0);
  add_integer_constant("GL_STENCIL_INDEX", 6401, 0);
  add_integer_constant("GL_STENCIL_PASS_DEPTH_FAIL", 2965, 0);
  add_integer_constant("GL_STENCIL_PASS_DEPTH_PASS", 2966, 0);
  add_integer_constant("GL_STENCIL_REF", 2967, 0);
  add_integer_constant("GL_STENCIL_TEST", 2960, 0);
  add_integer_constant("GL_STENCIL_VALUE_MASK", 2963, 0);
  add_integer_constant("GL_STENCIL_WRITEMASK", 2968, 0);
  add_integer_constant("GL_STEREO", 3123, 0);
  add_integer_constant("GL_SUBPIXEL_BITS", 3408, 0);
  add_integer_constant("GL_SUNX_geometry_compression", 1, 0);
  add_integer_constant("GL_SUNX_surface_hint", 1, 0);
  add_integer_constant("GL_SUN_convolution_border_modes", 1, 0);
  add_integer_constant("GL_SUN_multi_draw_arrays", 1, 0);
  add_integer_constant("GL_SURFACE_SIZE_HINT_SUNX", 33234, 0);
  add_integer_constant("GL_T", 8193, 0);
  add_integer_constant("GL_T2F_C3F_V3F", 10794, 0);
  add_integer_constant("GL_T2F_C4F_N3F_V3F", 10796, 0);
  add_integer_constant("GL_T2F_C4UB_V3F", 10793, 0);
  add_integer_constant("GL_T2F_N3F_V3F", 10795, 0);
  add_integer_constant("GL_T2F_V3F", 10791, 0);
  add_integer_constant("GL_T4F_C4F_N3F_V4F", 10797, 0);
  add_integer_constant("GL_T4F_V4F", 10792, 0);
  add_integer_constant("GL_TABLE_TOO_LARGE_EXT", 32817, 0);
  add_integer_constant("GL_TEXTURE", 5890, 0);
  add_integer_constant("GL_TEXTURE_1D", 3552, 0);
  add_integer_constant("GL_TEXTURE_2D", 3553, 0);
  add_integer_constant("GL_TEXTURE_3D_EXT", 32879, 0);
  add_integer_constant("GL_TEXTURE_ALPHA_SIZE", 32863, 0);
  add_integer_constant("GL_TEXTURE_BINDING_1D", 32872, 0);
  add_integer_constant("GL_TEXTURE_BINDING_2D", 32873, 0);
  add_integer_constant("GL_TEXTURE_BIT", 262144, 0);
  add_integer_constant("GL_TEXTURE_BLUE_SIZE", 32862, 0);
  add_integer_constant("GL_TEXTURE_BORDER", 4101, 0);
  add_integer_constant("GL_TEXTURE_BORDER_COLOR", 4100, 0);
  add_integer_constant("GL_TEXTURE_COLOR_TABLE_SGI", 32956, 0);
  add_integer_constant("GL_TEXTURE_COMPONENTS", 4099, 0);
  add_integer_constant("GL_TEXTURE_COORD_ARRAY", 32888, 0);
  add_integer_constant("GL_TEXTURE_COORD_ARRAY_POINTER", 32914, 0);
  add_integer_constant("GL_TEXTURE_COORD_ARRAY_SIZE", 32904, 0);
  add_integer_constant("GL_TEXTURE_COORD_ARRAY_STRIDE", 32906, 0);
  add_integer_constant("GL_TEXTURE_COORD_ARRAY_TYPE", 32905, 0);
  add_integer_constant("GL_TEXTURE_DEPTH_EXT", 32881, 0);
  add_integer_constant("GL_TEXTURE_ENV", 8960, 0);
  add_integer_constant("GL_TEXTURE_ENV_COLOR", 8705, 0);
  add_integer_constant("GL_TEXTURE_ENV_MODE", 8704, 0);
  add_integer_constant("GL_TEXTURE_GEN_MODE", 9472, 0);
  add_integer_constant("GL_TEXTURE_GEN_Q", 3171, 0);
  add_integer_constant("GL_TEXTURE_GEN_R", 3170, 0);
  add_integer_constant("GL_TEXTURE_GEN_S", 3168, 0);
  add_integer_constant("GL_TEXTURE_GEN_T", 3169, 0);
  add_integer_constant("GL_TEXTURE_GREEN_SIZE", 32861, 0);
  add_integer_constant("GL_TEXTURE_HEIGHT", 4097, 0);
  add_integer_constant("GL_TEXTURE_INTENSITY_SIZE", 32865, 0);
  add_integer_constant("GL_TEXTURE_INTERNAL_FORMAT", 4099, 0);
  add_integer_constant("GL_TEXTURE_LUMINANCE_SIZE", 32864, 0);
  add_integer_constant("GL_TEXTURE_MAG_FILTER", 10240, 0);
  add_integer_constant("GL_TEXTURE_MATRIX", 2984, 0);
  add_integer_constant("GL_TEXTURE_MIN_FILTER", 10241, 0);
  add_integer_constant("GL_TEXTURE_PRIORITY", 32870, 0);
  add_integer_constant("GL_TEXTURE_RED_SIZE", 32860, 0);
  add_integer_constant("GL_TEXTURE_RESIDENT", 32871, 0);
  add_integer_constant("GL_TEXTURE_STACK_DEPTH", 2981, 0);
  add_integer_constant("GL_TEXTURE_WIDTH", 4096, 0);
  add_integer_constant("GL_TEXTURE_WRAP_R_EXT", 32882, 0);
  add_integer_constant("GL_TEXTURE_WRAP_S", 10242, 0);
  add_integer_constant("GL_TEXTURE_WRAP_T", 10243, 0);
  add_integer_constant("GL_TRANSFORM_BIT", 4096, 0);
  add_integer_constant("GL_TRIANGLES", 4, 0);
  add_integer_constant("GL_TRIANGLE_FAN", 6, 0);
  add_integer_constant("GL_TRIANGLE_STRIP", 5, 0);
  add_integer_constant("GL_TRUE", 1, 0);
  add_integer_constant("GL_UNPACK_ALIGNMENT", 3317, 0);
  add_integer_constant("GL_UNPACK_IMAGE_HEIGHT_EXT", 32878, 0);
  add_integer_constant("GL_UNPACK_LSB_FIRST", 3313, 0);
  add_integer_constant("GL_UNPACK_ROW_LENGTH", 3314, 0);
  add_integer_constant("GL_UNPACK_SKIP_IMAGES_EXT", 32877, 0);
  add_integer_constant("GL_UNPACK_SKIP_PIXELS", 3316, 0);
  add_integer_constant("GL_UNPACK_SKIP_ROWS", 3315, 0);
  add_integer_constant("GL_UNPACK_SWAP_BYTES", 3312, 0);
  add_integer_constant("GL_UNSIGNED_BYTE", 5121, 0);
  add_integer_constant("GL_UNSIGNED_INT", 5125, 0);
  add_integer_constant("GL_UNSIGNED_SHORT", 5123, 0);
  add_integer_constant("GL_V2F", 10784, 0);
  add_integer_constant("GL_V3F", 10785, 0);
  add_integer_constant("GL_VENDOR", 7936, 0);
  add_integer_constant("GL_VERSION", 7938, 0);
  add_integer_constant("GL_VERSION_1_1", 1, 0);
  add_integer_constant("GL_VERTEX_ARRAY", 32884, 0);
  add_integer_constant("GL_VERTEX_ARRAY_POINTER", 32910, 0);
  add_integer_constant("GL_VERTEX_ARRAY_SIZE", 32890, 0);
  add_integer_constant("GL_VERTEX_ARRAY_STRIDE", 32892, 0);
  add_integer_constant("GL_VERTEX_ARRAY_TYPE", 32891, 0);
  add_integer_constant("GL_VIEWPORT", 2978, 0);
  add_integer_constant("GL_VIEWPORT_BIT", 2048, 0);
  add_integer_constant("GL_WRAP_BORDER_SUN", 33236, 0);
  add_integer_constant("GL_XOR", 5382, 0);
  add_integer_constant("GL_ZERO", 0, 0);
  add_integer_constant("GL_ZOOM_X", 3350, 0);
  add_integer_constant("GL_ZOOM_Y", 3351, 0);
  post_init();
}


